"""A toy hackathon project for gpt-powered workflows.
"""

import inspect
import json
import os
from typing import Callable, List, Type

import docstring_parser
import joblib
import openai
import pandas as pd
from sklearn.linear_model import LogisticRegression

from flytekit import current_context, task, Deck, ImageSpec, Secret
from flytekit.configuration import Config, PlatformConfig
from flytekit.remote import FlyteRemote
from flytekit.types.file import FlyteFile
from flytekit.experimental import eager

from auto_workflow import auto


image_spec = ImageSpec(
    "auto-workflows",
    registry="ghcr.io/unionai-oss",
    requirements="requirements.txt",
    python_version="3.10",
)


CLIENT_SECRET_GROUP = "arn:aws:secretsmanager:us-east-2:356633062068:secret:"
CLIENT_SECRET_KEY = "eager-workflow-LhHnqo"
OPENAI_API_SECRET_GROUP = "arn:aws:secretsmanager:us-east-2:356633062068:secret:"
OPENAI_API_SECRET_KEY = "openai_secret_key-1QGSZb"


@task(container_image=image_spec)
def add_numbers(x: float, y: float) -> FlyteFile:
    """Gets the sum of two numbers.
    
    :param x: The first number.
    :param y: The second number.
    """
    print(f"Adding numbers {x} and {y}")
    out = x + y
    output_path = os.path.join(current_context().working_directory, "result.txt")
    with open(output_path, "w") as f:
        f.write(str(out))
    return FlyteFile(output_path)


@task(container_image=image_spec)
def multiply_numbers(x: float, y: float) -> FlyteFile:
    """Gets the product of two numbers.
    
    :param x: The first number.
    :param y: The second number.
    """
    print(f"Multiplying numbers {x} and {y}")
    out = x * y
    output_path = os.path.join(current_context().working_directory, "result.txt")
    with open(output_path, "w") as f:
        f.write(str(out))
    return FlyteFile(output_path)


@task(container_image=image_spec)
def concat_strings(strings: List[str]) -> FlyteFile:
    """Concatenates a list of strings into one.
    
    :param strings: A list of strings to concatenate
    """
    print(f"Concatenating strings {strings}")
    result = "".join(strings)
    output_path = os.path.join(current_context().working_directory, "result.txt")
    with open(output_path, "w") as f:
        f.write(result)
    return FlyteFile(output_path)


@task(container_image=image_spec)
def train_classifier(
    training_data: List[dict],
    target_column: str,
) -> FlyteFile:
    """Trains a classifier on the training data.
    
    :param training_data: A list of dictionaries, where each dictionary is a row of the training data.
    :param target_column: The name of the column to use as the target.
    """
    print(f"Training a classifier on training data: {training_data[:3]}")
    training_data = pd.DataFrame.from_records(training_data)
    features, target = training_data.drop(target_column, axis="columns"), training_data[target_column]
    model = LogisticRegression(max_iter=10_000)
    model.fit(features, target)
    output_path = os.path.join(current_context().working_directory, "result.pkl")
    joblib.dump(model, output_path)
    return FlyteFile(output_path)


remote = FlyteRemote(
    config=Config(
        platform=PlatformConfig(
            endpoint="demo.hosted.unionai.cloud",
            client_id="eager-workflows",
        ),
    ),
    default_project="flytesnacks",
    default_domain="development",
)


@autoflow(
    remote=remote,
    client_secret_group=CLIENT_SECRET_GROUP,
    client_secret_key=CLIENT_SECRET_KEY,
    tasks=[add_numbers, multiply_numbers, concat_strings, train_classifier],
    model="gpt-3.5-turbo-1106",
    openai_secret_group=OPENAI_API_SECRET_GROUP,
    openai_secret_key=OPENAI_API_SECRET_KEY,
    container_image=image_spec,
    local_entrypoint=False,
)
async def auto_wf(prompt: str, inputs: dict) -> FlyteFile:
    """You are a helpful bot that picks functions based on a prompt and a set of inputs.

    What tool should I use for completing the task '{prompt}' using the following inputs?
    {inputs}
    """



"""GPT-powered workflows."""

import inspect
import os
from functools import partial, wraps
from typing import Callable, Type

import docstring_parser
import openai

from flytekit import current_context, Deck, Secret
from flytekit.experimental import eager


def define_tool(fn: Callable) -> dict:
    docstring = docstring_parser.parse(fn.__doc__)
    description = (
        docstring.short_description
        if docstring.long_description is None
        else f"{docstring.short_description}\n\n{docstring.long_description}"
    )
    signature = inspect.signature(fn)

    def get_json_datatype(type: Type):
        if type is str:
            return "string"
        if type in (int, float):
            return "number"
        return "object"

    return {
        "type": "function",
        "function": {
            "name": fn.__name__,
            "description": description,
            "parameters": {
                "type": "object",
                "properties": {
                    p.arg_name: {
                        "type": get_json_datatype(fn.__annotations__[p.arg_name]),
                        "description": p.description,
                    }
                    for p in docstring.params
                },
                "required": [
                    p.arg_name for p in docstring.params
                    if signature.parameters[p.arg_name].default is inspect._empty
                ],
            }
        }
    }


DECK_HTML_TEMPLATE = """
<h2>Auto-workflow Output</h2>
<hr>
<p>{output_contents}</p>
"""

def autoflow(
    _fn=None,
    *,
    tasks,
    remote,
    client_secret_group,
    client_secret_key,
    model,
    openai_secret_group,
    openai_secret_key,
    **kwargs,
):
    
    if _fn is None:
        return partial(
            autoflow,
            tasks=tasks,
            remote=remote,
            client_secret_group=client_secret_group,
            client_secret_key=client_secret_key,
            model=model,
            openai_secret_group=openai_secret_group,
            openai_secret_key=openai_secret_key,
            **kwargs,
        )

    fn_docstring = docstring_parser.parse(_fn.__doc__)
    secret_requests = kwargs.pop("secret_requests", None) or []
    secret_requests.append(Secret(group=openai_secret_group, key=openai_secret_key))
    
    @wraps(_fn)
    async def wrapper(prompt: str, inputs: dict):
        
        try:
            os.environ["OPENAI_API_KEY"] = current_context().secrets.get(
                openai_secret_group,
                openai_secret_key,
            )
        except ValueError:
            pass

        await _fn(prompt, inputs)

        client = openai.OpenAI()

        _globals = wrapper.__globals__
        available_tasks = {task.task_function.__name__: _globals[task.task_function.__name__] for task in tasks}
        tools = [define_tool(task.entity.task_function) for task in available_tasks.values()]

        messages = [
            {"role": "system", "content": fn_docstring.short_description},
            {"role": "user", "content": fn_docstring.long_description.format(prompt=prompt, inputs=inputs)},
        ]
        
        print("Getting task choice:")
        print(f"🧰 Tools:")
        for tool in tools:
            print(f"  - {tool}")

        print(f"💬 Messages: {messages}")
        for message in messages:
            print(f"  - {message}")

        response = client.chat.completions.create(
            model=model,
            messages=messages,
            tools=tools,
            tool_choice="auto",
        )
        response_message = response.choices[0].message

        # just use the first tool chosen
        tool_call = response_message.tool_calls[0]
        task_choice = available_tasks[tool_call.function.name]
        print(f"👉 Task choice: {tool_call.function.name}")
        print(f"🔢 Inputs: {inputs}")
        result = await task_choice(**inputs)

        result.download()
        try:
            with open(result.path) as f:
                output_contents = f.read()
        except Exception:
            with open(result.path, "rb") as f:
                output_contents = f.read()
        
        Deck("auto workflow output", DECK_HTML_TEMPLATE.format(output_contents=output_contents))
        return result
    
    for t in tasks:
        # hack: this makes sure the wrapper function has access to the task functions in
        # their AsyncEntity form.
        wrapper.__globals__[t.__name__] = _fn.__globals__[t.__name__]
    
    return eager(
        remote=remote,
        client_secret_group=client_secret_group,
        client_secret_key=client_secret_key,
        secret_requests=secret_requests,
        **kwargs
    )(wrapper)



import pathlib

from setuptools import find_packages, setup

__version__ = "0.0.1"


with open("requirements.txt") as f:
    install_requires = [x.strip() for x in f.readlines()]

LICENSE: str = "Apache"
README: str = pathlib.Path("README.md").read_text()

setup(
    name="autoflows",
    version=__version__,
    author="unionai-oss",
    author_email="info@union.ai",
    description="Automated Flyte workflows, powered by GPTs.",
    long_description=README,
    long_description_content_type="text/markdown",
    license=LICENSE,
    keywords=["machine-learning", "artificial-intelligence", "orchestration"],
    data_files=[("", ["LICENSE"])],
    include_package_data=True,
    packages=find_packages(
        include=["autoflows*"],
        exclude=["tests*"],
    ),
    package_data={"autoflows": ["py.typed"]},
    python_requires=">3.7",
    platforms="any",
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        f"License :: OSI Approved :: {LICENSE} Software License",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    url="https://github.com/unionai-oss/autoflows/",
    project_urls={
        "Source Code": "https://github.com/unionai-oss/autoflows/",
        "Issue Tracker": "https://github.com/unionai-oss/autoflows/issues",
    },
)
